

set -ex



pip check
transformers-cli --help
pytest tests -k "not (return_timestamps_ctc_fast or chunking_fast or small_model_pt or from_pretrained_low_cpu_mem_usage_measured or test_torch_compile_fullgraph or test_word_heuristic_leading_space)"  --ignore=tests/agents --ignore=tests/benchmark --ignore=tests/bettertransformer --ignore=tests/deepspeed --ignore=tests/extended --ignore=tests/fixtures --ignore=tests/fsdp --ignore=tests/generation --ignore=tests/models --ignore=tests/optimization --ignore=tests/peft_integration --ignore=tests/pipelines --ignore=tests/quantization --ignore=tests/repo_utils --ignore=tests/sagemaker --ignore=tests/tokenization --ignore=tests/trainer --ignore=tests/utils
exit 0
